##############################################################################
#
# Copyright (c) 2010 ViFiB SARL and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import hashlib
import httplib
import json
import os
import socket
import subprocess
import tempfile
import traceback
import urllib2
import urlparse

# XXX: code between select/select_generic must be factored

# Timeout here is about timeout to CONNECT to the server (socket initialization then server answers actual data), not to retrieve/send informations.
# To be clear: it is NOT about uploading/downloading data, but about time to connect to the server, then time that server takes to start answering.
TIMEOUT = 60
# Same here. We just wait longer that, after having uploaded the file, the server digests it. It can take time.
UPLOAD_TIMEOUT = 60 * 60


def urljoin(a, b):
  if not a.endswith('/'):
    a += '/'
  return a + b


class hashing_file(object):

  def __init__(self, file):
    self._f = file
    self._h = hashlib.sha512()

  def hexdigest(self):
    assert not self._f.read(1)
    return self._h.hexdigest()

  def read(self, *args, **kw):
    d = self._f.read(*args, **kw)
    self._h.update(d)
    return d

  def __len__(self):
    f = self._f
    try:
      fd = f.fileno()
    except AttributeError:
      pos = f.tell()
      try:
        f.seek(0, 2)
        return f.tell()
      finally:
        f.seek(pos)
    return os.fstat(fd).st_size


class NetworkcacheClient(object):
  '''
    NetworkcacheClient is a wrapper for httplib.
    It must implement all the required methods to use:
     - SHADIR
     - SHACACHE
  '''

  openssl = 'openssl'
  def parseUrl(self, url):
    return_dict = {}
    parsed_url = urlparse.urlparse(url)
    return_dict['header_dict'] = {'Content-Type': 'application/json'}
    user = parsed_url.username
    passwd = parsed_url.password
    if user is not None:
      return_dict['header_dict']['Authorization'] = 'Basic %s' %\
        ('%s:%s' % (user, passwd)).encode('base64').strip()

    return_dict['path'] = parsed_url.path
    return_dict['host'] = parsed_url.hostname
    return_dict['scheme'] = parsed_url.scheme
    return_dict['port'] = parsed_url.port or \
                           socket.getservbyname(parsed_url.scheme)

    return return_dict

  def __init__(self, shacache, shadir, signature_private_key_file=None,
      signature_certificate_list=None, shacache_key_file=None,
      shacache_cert_file=None, shadir_key_file=None, shadir_cert_file=None):
    """Initializes shacache object.

    Parameters:
      shacache
        URL to shacache.
        Required.

      shadir
        URL to shadir.
        Required.

      signature_private_key_file
        Path to private key file used for signing content.
        Optional.

      signature_certificate_list
        List of strings of certificates to verify content.
        Optional

      shacache_key_file
        Key file used to authenticate to shacache.
        Optional.

      shacache_cert_file
        Certificate file used to authenticate to shacache.
        Optional.

      shadir_key_file
        Key file used to authenticate to shadir.
        Optional.

      shadir_cert_file
        Certificate file used to authenticate to shadir.
        Optional.
      """
    # ShaCache Properties
    for k, v in self.parseUrl(shacache).iteritems():
      setattr(self, 'shacache_%s' % k, v)
    self.shacache_url = shacache
    self.shadir_url = shadir

    # ShaDir Properties
    for k, v in self.parseUrl(shadir).iteritems():
      setattr(self, 'shadir_%s' % k, v)

    self.signature_private_key_file = signature_private_key_file
    if type(signature_certificate_list) is str:
      # If signature_certificate_list is a string, parse it to a list of
      # certificates
      cert_marker = "-----BEGIN CERTIFICATE-----"
      parsed_signature_certificate_list = [cert_marker + '\n' + q.strip() \
        for q in signature_certificate_list.split(cert_marker) \
          if q.strip()]
      self.signature_certificate_list = parsed_signature_certificate_list
    else:
      self.signature_certificate_list = signature_certificate_list

    self.shacache_key_file = shacache_key_file
    self.shacache_cert_file = shacache_cert_file
    self.shadir_key_file = shadir_key_file
    self.shadir_cert_file = shadir_cert_file

  def upload(self, file_descriptor, key=None, urlmd5=None, file_name=None,
                                    valid_until=None, architecture=None, **kw):
    ''' Upload the file to the server.
    If key is None it must only upload to SHACACHE.
    Otherwise, it must create a new entry on SHADIR.
    '''
    # do not trust, go to beginning of opened file
    file_descriptor.seek(0)
    file_descriptor = hashing_file(file_descriptor)
    if self.shacache_scheme == 'https':
      shacache_connection = httplib.HTTPSConnection(self.shacache_host,
        self.shacache_port, key_file=self.shacache_key_file,
        cert_file=self.shacache_cert_file, timeout=UPLOAD_TIMEOUT)
    else:
      shacache_connection = httplib.HTTPConnection(self.shacache_host,
        self.shacache_port, timeout=UPLOAD_TIMEOUT)
    try:
      shacache_connection.request('POST', self.shacache_path, file_descriptor,
                                      self.shacache_header_dict)
      result = shacache_connection.getresponse()
      sha512sum = result.read()
    finally:
      shacache_connection.close()

    if result.status != 201 or sha512sum != file_descriptor.hexdigest():
      raise UploadError('Failed to upload the file to SHACACHE Server.'
                        'URL: %s. Response code: %s. Response data: %s'
                        % (self.shacache_host, result.status, sha512sum))

    if key is not None:
      kw['sha512'] = sha512sum # always update sha512sum
      if file_name is None or urlmd5 is None:
        raise ValueError('file_name and urlmd5 are required'
                         ' for non-generic upload')
      if valid_until is not None:
        kw['valid-until'] = valid_until
      if architecture is not None:
        kw['architecture'] = architecture
      self.index(key, file=file_name, urlmd5=urlmd5, **kw)

    return sha512sum

  upload_generic = upload # BBB

  def index(self, key, **kw):
    data = json.dumps(kw)
    try:
      data = [data, self._getSignatureString(data)]
    except Exception:
      raise UploadError('Impossible to sign content, error:\n%s'
                        % traceback.format_exc())
    if self.shadir_scheme == 'https':
      shadir_connection = httplib.HTTPSConnection(self.shadir_host,
          self.shadir_port, key_file=self.shadir_key_file,
          cert_file=self.shadir_cert_file, timeout=UPLOAD_TIMEOUT)
    else:
      shadir_connection = httplib.HTTPConnection(self.shadir_host,
          self.shadir_port, timeout=UPLOAD_TIMEOUT)
    try:
      shadir_connection.request('PUT', '/'.join([self.shadir_path, key]),
          json.dumps(data), self.shadir_header_dict)
      result = shadir_connection.getresponse()
      data = result.read()
    finally:
      shadir_connection.close()

    if result.status != 201:
      raise UploadError('Failed to upload data to SHADIR Server.'
                        'URL: %s. Response code: %s. Response data: %s'
                        % (self.shadir_host, result.status, data))

  def download(self, sha512sum):
    ''' Download the file.
    It uses http GET request method.
    # do not trust, go to beginning of opened file
    file_descriptor.seek(0)
    file_descriptor = hashing_file(file_descriptor)
    '''
    sha_cache_url = urljoin(self.shacache_url, sha512sum)
    request = urllib2.Request(url=sha_cache_url, data=None,
      headers=self.shadir_header_dict)
    return urllib2.urlopen(request, timeout=TIMEOUT)

  def select(self, key):
    ''' Download a file from shacache by selecting the entry in shadir
    Raise DirectoryNotFound if no trustable file is found.
    '''
    url = urljoin(self.shadir_url, key)
    request = urllib2.Request(url=url, data=None,
      headers=self.shadir_header_dict)
    data = urllib2.urlopen(request, timeout=TIMEOUT).read()
    # Filtering...
    try:
      data_list = json.loads(data)
    except Exception:
      raise DirectoryNotFound('It was impossible to parse json response:\n%s'%
        traceback.format_exc())
    filtered_data_list = []
    if self.signature_certificate_list is not None:
      for data in data_list:
        if len(data[1]):
          if self._verifySignatureInCertificateList(data[0], data[1]):
            filtered_data_list.append(data)
    else:
      filtered_data_list = data_list

    if len(filtered_data_list) == 0:
      raise DirectoryNotFound('Could not find a trustable entry.')

    information_json, signature = filtered_data_list[0]
    try:
      information_dict = json.loads(information_json)
    except Exception:
      raise DirectoryNotFound('It was impossible to parse json-in-json '
        'response:\n%s' % traceback.format_exc())
    try:
      sha512 = information_dict.get('sha512')
    except Exception:
      raise DirectoryNotFound('It was impossible to fetch sha512 from '
        'directory response (%r):\n%s' % (information_dict,
          traceback.format_exc()))
    return self.download(sha512)

  def select_generic(self, key):
    ''' Select trustable entries from shadir.
    '''
    url = urljoin(self.shadir_url, key)
    request = urllib2.Request(url=url, data=None,
      headers=self.shadir_header_dict)
    data = urllib2.urlopen(request, timeout=TIMEOUT).read()
    try:
      data_list = json.loads(data)
    except Exception:
      raise DirectoryNotFound('It was impossible to parse json response:\n%s' %
        traceback.format_exc())
    filtered_data_list = []
    for data in data_list:
      if len(data[1]):
        if self._verifySignatureInCertificateList(data[0], data[1]):
          filtered_data_list.append(data)
    return filtered_data_list

  def _openssl(self, input, *args):
    p = subprocess.Popen((self.openssl,) + args,
                         stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    output = p.communicate(input)[0]
    if p.returncode:
      raise subprocess.CalledProcessError(p.returncode, self.openssl, output)
    return output

  def _getSignatureString(self, content):
    """
      Return the signature based on certification file.
    """
    if self.signature_private_key_file is None:
      return ''
    return self._openssl(content, "dgst", "-sha1", "-sign",
      self.signature_private_key_file).encode('base64')

  def _verifySignatureInCertificateList(self, content, signature_string):
    """
      Returns true if it can find any valid certificate or false if it does not
      find any.
    """
    if self.signature_certificate_list:
      with tempfile.NamedTemporaryFile() as signature_file:
        signature_file.write(signature_string.decode('base64'))
        signature_file.flush()
        for certificate in self.signature_certificate_list:
          try:
            pubkey = self._openssl(certificate, "x509", "-pubkey", "-noout")
            with tempfile.NamedTemporaryFile() as pubkey_file:
              pubkey_file.write(pubkey)
              pubkey_file.flush()
              if self._openssl(content, "dgst", "-sha1", "-verify",
                  pubkey_file.name, "-signature", signature_file.name
                  ).startswith('Verified OK'):
                return True
          except Exception:
            # in case of failure, emit *anything*, but swallow all what possible
            traceback.print_exc()
    return False


class DirectoryNotFound(Exception):
  pass


class UploadError(Exception):
  pass

